from paradoxdjango.db import connection
from paradoxdjango.test import TestCase


class SchemaLoggerTests(TestCase):
    def test_extra_args(self):
        editor = connection.schema_editor(collect_sql=True)
        sql = "SELECT * FROM foo WHERE id in (%s, %s)"
        params = [42, 1337]
        with self.assertLogs("paradoxdjango.db.backends.schema", "DEBUG") as cm:
            editor.execute(sql, params)
        self.assertEqual(cm.records[0].sql, sql)
        self.assertEqual(cm.records[0].params, params)
        self.assertEqual(
            cm.records[0].getMessage(),
            "SELECT * FROM foo WHERE id in (%s, %s); (params [42, 1337])",
        )
